import BlockKind;

/*
 *
 * Block
 *
 */

public class Block 
{
    protected BlockKind m_kind;
	protected Block m_branch;
	protected Block m_fail;
	protected Block m_link;
    protected Block m_prev;
    protected Block m_next;
	protected Node m_first;
	protected Node m_last;
    protected Block m_dominatorChild;
    protected Block m_dominatorNext;
    protected int m_number;
    private boolean m_visiting;

	public static Block startBlock = null;
	public static Block endBlock = null;

    protected static int m_count;


	public Block()
	{
        m_kind = BlockKind.SEQUENCE;
        m_branch = null;
		m_fail = null;
		m_link = null;
		m_first = m_last = null;
        m_next = m_prev = null;
        m_dominatorChild = m_dominatorNext = null;
        m_number = m_count++;
        m_visiting = false;
	}

	public Block( BlockKind kind )
	{
        m_kind = kind;
        m_branch = null;
		m_fail = null;
		m_link = null;
		m_first = m_last = null;
        m_next = m_prev = null;
        m_dominatorChild = m_dominatorNext = null;
        m_number = m_count++;
        m_visiting = false;
	}

    public void SetKind( BlockKind kind )
    {
        m_kind = kind;
    }

    public BlockKind GetKind()
    {
        return m_kind;
    }

    public void SetBranch( Block branch )
    {
        m_branch = branch;
    }

    public Block GetBranch()
    {
        return m_branch;
    }

    public Block GetFail()
    {
        return m_fail;
    }

    public void SetFail( Block fail )
    {
        m_fail = fail;
    }

    public Node GetFirst()
    {
        return m_first;
    }

    public void SetFirst( Node first )
    {
        m_first = first;
    }

    public Node GetLast()
    {
        return m_last;
    }

    public void SetLast( Node last )
    {
        m_last = last;
    }

    public Block GetLink()
    {
        return m_link;
    }

    public void SetLink( Block link )
    {
        m_link = link;
    }

    public Block GetChildDominator()
	{
		return m_dominatorChild;
	}

    public Block GetSiblingDominator()
	{
		return m_dominatorNext;
	}

    public void SetDominator( Block b )
    {
        if( b.m_dominatorChild == null )
            b.m_dominatorChild = this;
        else
        {
            Block d = b.m_dominatorChild;
            while( d.m_dominatorNext != null )
                d = d.m_dominatorNext;
            d.m_dominatorNext = this;
        }
    }

    public Block EndBlock( Node e )
    {
        if( m_first == null )
            m_first = e;
        m_last = e;
        return this;
    }

	public void AddTail()
	{
		if( endBlock != null )
		{
			endBlock.m_next = this;
			m_prev = endBlock;
			m_next = null;
			endBlock = this;
		}
		else
		{
			startBlock = endBlock = this;
		}
	}


	public String Traverse()
	{
		String str = "No Block Tree\n";
		if( startBlock != null )
		{
			str =  "Block Tree\n";
            str += "==========\n";
			str += startBlock.toString( 2 );
		}

		return str;
	}

	public String TraverseDominatorTree()
	{
		String str = "No Dominator Tree\n";
		if( startBlock != null )
		{
			str =  "Dominator Tree\n";
            str += "==============\n";
			str += startBlock.dominatorToString( 2 );
		}

		return str;
	}

	public String toString( int ind )
	{
		String str = indent( ind );
		str += "Block " + m_number + "\n";
        if( m_visiting )
            return str;
        m_visiting = true;
        if( m_first != null )
        {
            Node node = m_first;
            do
            {
    			str += node.toString( ind+2 );
                if( node == m_last )
                    break;
                node = node.GetNext();
            }
            while( node != null );
		}
		if( m_branch != null )
			str += indent( ind ) + "Branch:\n" + m_branch.toString( ind+2 );
		if( m_fail != null )
			str += indent( ind ) + "Fail:\n" + m_fail.toString( ind+2 );
        m_visiting = false;
		return str;
	}

	public String toString()
	{
		String str = "Block " + m_number;
		return str;
	}

	private String dominatorToString( int ind )
	{
		String str = indent( ind );
		str += this + "\n";
		if( m_dominatorChild != null )
			str += m_dominatorChild.dominatorToString( ind + 2 );
        if( m_dominatorNext != null )
			str += m_dominatorNext.dominatorToString( ind );
		return str;
	}

	private String indent( int ind )
	{
		String str = "";
		for( int i=0; i<ind; i++ )
			str += " ";
		return str;
	}
}

