import java.util.*;
import java.io.*;
import Node;
import Block;

/*
 *
 * ParseTree
 *
 */
public class ParseTree 
{
    protected Hashtable m_symbolTable;        // holds last use of symbols
    protected Node m_nodeRoot;                // root of node list
    protected Block m_blockRoot;              // root of block list
    protected Block m_currentBlock;           // refers to current block
    protected PrintStream m_ps;
    protected int m_Assignments;
    protected boolean m_firstPhi;

    public ParseTree( PrintStream ps )
    {
        m_symbolTable = new Hashtable();
        m_nodeRoot = null;
        m_blockRoot = null;
        m_currentBlock = null;
        m_ps = ps;
        InitNodes();
        m_Assignments = 2;
        m_firstPhi = true;
    }

    // Non-Instruction nodes

    // save number in symbol table if not already in there
	public Node Number( int n )
	{
        return Number( new Integer(n) );
    }

	public Node Number( Integer n )
	{
		Node newNode = null;
        newNode = (Node)m_symbolTable.get( n.toString() );
		if( newNode == null )
		{
			newNode = new Node( GetCurrentBlock() );
			newNode.SetKind(  NodeKind.INTEGER );
			newNode.SetValue( n.toString() );
			newNode.AddTail(); // add to global node list
			m_symbolTable.put( n.toString(), newNode );
		}
		return newNode;
	}

    // create new name node
    public Node Ident( String id, boolean write )
	{
		Node newNode = null;
        if( write )
        {
            // create new node, save reference to last use
            // therefore, this reference is used in later reads
		    newNode = new Node( GetCurrentBlock() );
		    newNode.SetKind( NodeKind.ID );
		    newNode.SetValue( id );
            newNode.SetlValue();
		    newNode.SetLastUse( (Node)m_symbolTable.get( id ) );
    		newNode.AddTail();
    		m_symbolTable.put( id, newNode );
        }
        else
        {
            // save node in table if not already there
            newNode = (Node)m_symbolTable.get( id );
		    if( newNode == null )
		    {
		        newNode = new Node( GetCurrentBlock() );
		        newNode.SetKind( NodeKind.ID );
		        newNode.SetValue( id );
        		newNode.AddTail();
        		m_symbolTable.put( id, newNode );
            }
        }
		return newNode;
	}

    public Node Ident( String id1, String id2, boolean write )
	{
        return Ident( id1 + id2, write );
	}

    // assign const expression to identifier, ie. CONST ident = expression
	public Node ConstDeclaration( String id, Node e )
	{
        Node resultNode = new Node( GetCurrentBlock() );
		resultNode.SetKind( NodeKind.CONST );
        resultNode.SetValue( id );
        resultNode.SetX( Ident( id, true ) );
        resultNode.SetY( e );
        resultNode.AddTail();
        return resultNode;
	}

    // type declaration, ie. TYPE ident=type
	public Node TypeDeclaration( String id, Vector typelist )
	{
        Node resultNode = null;
        for( int i=0; i<typelist.size(); i++ )
		{
            resultNode = new Node( GetCurrentBlock() );
			resultNode.SetKind( NodeKind.TYPE );
            Node t = (Node)typelist.elementAt( i );
            if( t.GetKind() == NodeKind.ID )
            {
                resultNode.SetValue( id );
                resultNode.SetX( Ident( id, true ) );
                resultNode.SetY( t );
            }
            else if( t.GetKind() == NodeKind.ARRAY )
            {
                resultNode.SetValue( id );
                resultNode.SetY( t );
            }
            else if( t.GetKind() == NodeKind.RECORD )
            {
                resultNode.SetX( Ident( id, false ) );
                resultNode.SetY( t );
                resultNode.SetValue( id );
		        resultNode.SetLastUse( (Node)m_symbolTable.get( resultNode.GetValue() ) );
            }
            resultNode.AddTail();
    		m_symbolTable.put( resultNode.GetValue(), resultNode );
		}
        return resultNode;
	}

    // variable declaration, assigns types to names
    // ie. VAR ident, ident : type
	public Node VarDeclaration( Vector nodelist, Vector typelist )
	{
        Node type = null;
        Node resultNode = null;
        for( int i=0; i<nodelist.size(); i++ )
		{
            for( int j=0; j<typelist.size(); j++ )
            {
                resultNode = new Node( GetCurrentBlock() );
			    resultNode.SetKind( NodeKind.VAR );
                resultNode.SetX( (Node)nodelist.elementAt( i ) );
                Node t = (Node)typelist.elementAt( j );
                resultNode.SetValue( resultNode.GetX().GetValue() );
                if( t.GetKind() == NodeKind.ID )
                {
                    resultNode.SetValue( resultNode.GetX().GetValue() );
                    resultNode.SetY( t );
		            resultNode.SetLastUse( (Node)m_symbolTable.get( resultNode.GetValue() ) );
    		        m_symbolTable.put( resultNode.GetValue(), resultNode );
                    resultNode.AddTail();
                }
                else if( t.GetKind() == NodeKind.ARRAY )
                {
                    resultNode.SetValue( resultNode.GetX().GetValue() );
                    resultNode.SetY( t );
		            resultNode.SetLastUse( (Node)m_symbolTable.get( resultNode.GetValue() ) );
    		        m_symbolTable.put( resultNode.GetValue(), resultNode );
                    resultNode.AddTail();
                }
                else if( t.GetKind() == NodeKind.RECORD )
                {
                    Node newtype = new Node( GetCurrentBlock() );
			        newtype.SetKind( NodeKind.TYPE );
                    newtype.SetX( type );
                    newtype.SetY( t );
                    newtype.AddTail();
                    type = newtype;
                }
                else
                {
                    resultNode.SetValue( resultNode.GetX().GetValue() );
                    resultNode.SetY( t );
		            resultNode.SetLastUse( (Node)m_symbolTable.get( resultNode.GetValue() ) );
    		        m_symbolTable.put( resultNode.GetValue(), resultNode );
                    resultNode.AddTail();
                }
            }
            if( type != null )
            {
                resultNode.SetValue( resultNode.GetX().GetValue() );
                resultNode.SetY( type );
		        resultNode.SetLastUse( (Node)m_symbolTable.get( resultNode.GetValue() ) );
    		    m_symbolTable.put( resultNode.GetValue(), resultNode );
                resultNode.AddTail();
            }
		}
        return resultNode;
	}

    // collects an array of identifiers for variable declarations
    // and record field declarations
    public Vector IdentList( String id, Vector nodelist )
	{
        Node node = Ident( id, false );
        nodelist.addElement( node );
        return nodelist;
	}

    // array type declaration, ie. ARRAY expression OF type
    public Vector ArrayType( Node e, Vector typelist )
    {
        Vector v = new Vector();
        Node resultNode = null;
        for( int i=0; i<typelist.size(); i++ )
		{
            resultNode = new Node( GetCurrentBlock() );
		    resultNode.SetKind( NodeKind.ARRAY );
            Node t = (Node)typelist.elementAt( i );
            resultNode.SetX( e );
            resultNode.SetY( t );
            resultNode.SetValue( t.GetValue() );
            resultNode.AddTail();
            v.addElement( resultNode );
		}
        return v;
    }

    // creates an internal array of record fields
    // the fields may themselves be arrays or record fields
    public Vector FieldList( Vector fieldlist, Vector nodelist, Vector typelist )
	{
        for( int i=0; i<nodelist.size(); i++ )
		{
            for( int j=0; j<typelist.size(); j++ )
            {
                Node node1 = new Node( GetCurrentBlock() );
			    node1.SetKind( NodeKind.RECORD );
                node1.SetX( (Node)nodelist.elementAt( i ) );
                Node t = (Node)typelist.elementAt( j );
                node1.SetY( t );
                node1.SetValue( node1.GetX().GetValue() );
		        node1.SetLastUse( (Node)m_symbolTable.get( node1.GetValue() ) );
    		    m_symbolTable.put( node1.GetValue(), node1 );
                node1.AddTail();
                fieldlist.addElement( node1 );
            }
		}
        return fieldlist;
	}


    // Instruction generating nodes

    // assign right node to left node
    public Node Assignment( Node l, Node r, NodeKind k )
    {
        Node resultNode = null;
        if( k == NodeKind.ID )
        {
            resultNode = new Node( GetCurrentBlock() );
            resultNode.SetKind( NodeKind.INSTRUCTION );
            resultNode.SetValue( l.GetValue() );
            resultNode.SetX( l );
            resultNode.SetY( r );
            resultNode.SetOp( Operator.MOVE );
            // for assignment 1 only
            if( m_Assignments == 1 )
                resultNode.SetLastUse( (Node)m_symbolTable.get( resultNode.GetValue() ) );
            else
            {
                // for assignment 2
                Node lastUse = (Node)m_symbolTable.get( resultNode.GetValue() );
                if( lastUse == null )
                    resultNode.SetSSANumber( 0 );
                else
                {
                    resultNode.SetLastUse( lastUse );
                    resultNode.SetSSANumber( lastUse.GetSSANumber() + 1 );
                }
                UpdatePhi( resultNode );
                // end change for assignment 2
            }

            m_symbolTable.put( resultNode.GetValue(), resultNode );
    		resultNode.AddTail();
        }
        else if( k == NodeKind.ARRAY )
        {
            Node node1 = new Node( GetCurrentBlock() );
            node1.SetKind( NodeKind.INSTRUCTION );
            node1.SetX( l );  // i[j]:=k, this is j
            node1.SetY( Number( 4 ) );
            node1.SetOp( Operator.MUL );
            node1.AddTail();
            Node node2 = new Node( GetCurrentBlock() );
            node2.SetKind( NodeKind.INSTRUCTION );
            node2.SetX( (Node)m_symbolTable.get( "FP" ) );
            node2.SetY( l.GetX() );  // i[j]:=k, this is base address of i
            node2.SetOp( Operator.ADD );
            node2.AddTail();
            Node node3 = new Node( GetCurrentBlock() );
            node3.SetKind( NodeKind.INSTRUCTION );
            node3.SetX( node1 );
            node3.SetY( node2 );
            node3.SetOp( Operator.ADDA );
            node3.AddTail();
            Node node4 = new Node( GetCurrentBlock() );
            node4.SetKind( NodeKind.INSTRUCTION );
            node4.SetValue( l.GetValue() );
            node4.SetX( node3 );
            node4.SetY( r );    // i[j]:=k, this is k
            node4.SetOp( Operator.STORE );
            // for assignment 1 only
            if( m_Assignments == 1 )
		        node4.SetLastUse( (Node)m_symbolTable.get( node4.GetValue() ) );
            else
            {
                // for assignment 2
                Node lastUse = (Node)m_symbolTable.get( node4.GetValue() );
                if( lastUse == null )
                    node4.SetSSANumber( 0 );
                else
                {
                    node4.SetLastUse( lastUse );
                    node4.SetSSANumber( lastUse.GetSSANumber() + 1 );
                }
                UpdatePhi( node4 );
                // end change for assignment 2
            }

            m_symbolTable.put( node4.GetValue(), node4 );
            node4.AddTail();
            resultNode = node4;
        }
        else if( k == NodeKind.RECORD )
        {
            resultNode = new Node( GetCurrentBlock() );
            resultNode.SetKind( NodeKind.INSTRUCTION );

            resultNode.SetValue( l.GetValue() );
            resultNode.SetX( l.GetY() ); // i.j:=k, this is j
            resultNode.SetY( r );   // i.j:=k, this is k
            resultNode.SetOp( Operator.MOVE );
            // for assignment 1 only
            if( m_Assignments == 1 )
            {
                resultNode.SetLastUse( (Node)m_symbolTable.get( resultNode.GetValue() ) );
                m_symbolTable.put( resultNode.GetValue(), resultNode );
            }
            else
            {
                // for assignment 2
                Node lastUse = (Node)m_symbolTable.get( resultNode.GetValue() );
                m_symbolTable.put( resultNode.GetValue(), resultNode );
                if( lastUse == null )
                    resultNode.SetSSANumber( 0 );
                else
                {
                    resultNode.SetLastUse( lastUse );
                    resultNode.SetSSANumber( lastUse.GetSSANumber() + 1 );
                }
                UpdatePhi( resultNode );
                // end change for assignment 2
            }

            resultNode.AddTail();
        }
        return resultNode;
    }

    // multiply, divide or mod two nodes
    public Node Term( Operator op, Node l, Node r )
    {
		Node resultNode = new Node( GetCurrentBlock() );
		resultNode.SetKind( NodeKind.INSTRUCTION );
        resultNode.SetX( l );
        resultNode.SetY( r );
        resultNode.SetOp( op );
        resultNode.AddTail();
		return resultNode;
    }

    // add or subtract two nodes (plus possible unary minus on first node)
    public Node Expression( Operator op, Node l, Node r, boolean minusSign )
    {
        Node resultNode = null;
        if( minusSign )
        {
            Node node1 = new Node( GetCurrentBlock() );
            node1.SetKind( NodeKind.INSTRUCTION );
            node1.SetX( l );
            node1.SetOp( Operator.NEG );
            node1.AddTail();
            if( r == null )
                resultNode = node1;
            else
            {
                Node node2 = new Node( GetCurrentBlock() );
                node2.SetKind( NodeKind.INSTRUCTION );
                node2.SetX( node1 );
                node2.SetY( r );
                node2.SetOp( op );
                node2.AddTail();
                resultNode = node2;
            }
        }
        else
        {
            if( r == null )
                resultNode = l;
            else
            {
                resultNode = new Node( GetCurrentBlock() );
                resultNode.SetKind( NodeKind.INSTRUCTION );
                resultNode.SetX( l );
                resultNode.SetY( r );
                resultNode.SetOp( op );
                resultNode.AddTail();
            }
        }
        return resultNode;
    }

    // combination of id and selector,
    // i.e. id, id.j, or id[j]
    // note that all variables are read only
    // (assignment is handled elsewhere)
    public Node Selector( String id )
    {
        // special case: id only
        return Ident( id, false );
    }

    public Node Selector( String id, Node r, boolean array, boolean write )
    {
        // cases id.j and id[j]
        Node resultNode = null;
        if( array )
        {
            if( write )
            {
                resultNode = Ident( id, false );//(Node)m_symbolTable.Get( id );
            }
            else
            {
                Node node1 = new Node( GetCurrentBlock() );
                node1.SetKind( NodeKind.INSTRUCTION );
                node1.SetX( r );  // k:=i[j], this is j
                node1.SetY( Number( 4 ) );
                node1.SetOp( Operator.MUL );
                node1.AddTail();
                Node node2 = new Node( GetCurrentBlock() );
                node2.SetKind( NodeKind.INSTRUCTION );
                node2.SetX( (Node)m_symbolTable.get( "FP" ) );
                node2.SetY( Ident( id, false ) );  // k:=i[j], this is base address of i
                node2.SetOp( Operator.ADD );
                node2.AddTail();
                Node node3 = new Node( GetCurrentBlock() );
                node3.SetKind( NodeKind.INSTRUCTION );
                node3.SetX( node1 );
                node3.SetY( node2 );
                node3.SetOp( Operator.ADDA );
                node3.AddTail();
                Node node4 = new Node( GetCurrentBlock() );
                node4.SetKind( NodeKind.INSTRUCTION );
                node4.SetValue( id );   // need the value for uselist later on
                node4.SetX( node3 );    // k:=i[j], this is loads result k
                node4.SetOp( Operator.LOAD );
                node4.AddTail();
                // for assignment 1 only
                if( m_Assignments == 1 )
		            node4.SetLastUse( (Node)m_symbolTable.get( node4.GetValue() ) );
                else
                {
                    // for assignment 2
                    Node lastUse = (Node)m_symbolTable.get( node4.GetValue() );
                    if( lastUse == null )
                        node4.SetSSANumber( 0 );
                    else
                    {
                        node4.SetLastUse( lastUse );
                        node4.SetSSANumber( lastUse.GetSSANumber() + 1 );
                    }
//                    UpdatePhi( node4 );
                    // end change for assignment 2
                }
                resultNode = node4;
            }
        }
        else
        {
            if( write )
            {
                resultNode = new Node( GetCurrentBlock() );
                resultNode.SetKind( NodeKind.RECORD );
                resultNode.SetValue( id );
                resultNode.SetX( Ident( id , false ) );
                Node f = resultNode.GetX();
                while( f != null && ( f.GetKind() != NodeKind.VAR && f.GetKind() != NodeKind.TYPE ) )
                    f = f.GetLastUse();
                if( f != null && f.GetKind() == NodeKind.VAR )
                    f = f.GetY();
                while( f != null && f.GetKind() == NodeKind.TYPE )
                {
                    if( f.GetY() != null && f.GetY().GetValue().equals( r.GetValue() ) )
                    {
                        resultNode.SetY( f.GetY() );
                        break;
                    }
                    f = f.GetX();
                }
            }
            else
            {
                resultNode = Ident( id , false );
                Node f = resultNode;
                while( f != null && f.GetKind() != NodeKind.VAR )
                    f = f.GetLastUse();
                f = f.GetY();
                while( f != null && f.GetKind() == NodeKind.TYPE )
                {
                    if( f.GetY().GetValue().equals( r.GetValue() ) )
                    {
                        break;
                    }
                    f = f.GetX();
                }
            }
        }
        return resultNode;
    }

    // comparison of two expressions, then prepare branch, fail and join
    // blocks for if-then-else-end
    public Node If( Operator op, Node l, Node r )
    {
        Node resultNode = null;
        // create new condition block
        // Note: don't set branch of current block yet, we need it later
        Block cond = new Block( BlockKind.IF );
        Block curBlock = GetCurrentBlock();
        cond.SetLink( curBlock );
        cond.SetDominator( curBlock );

        Node node1 = new Node( cond );
        node1.SetKind( NodeKind.INSTRUCTION );
        node1.SetX( l );    // i < j etc., this is i
        node1.SetY( r );    // i < j etc., this is j
        node1.SetOp( Operator.CMP );
        node1.AddTail();
        Node node2 = new Node( cond );
        node2.SetKind( NodeKind.INSTRUCTION );
        node2.SetX( node1 );
        node2.SetOp( op );  // blt, etc to y, i.e. to branch block
        node2.AddTail();
        Node node3 = new Node( cond );
        node3.SetKind( NodeKind.INSTRUCTION );
        node3.SetOp( Operator.BRA );    // bra to y, i.e. to fail block
        node3.AddTail();
        resultNode = node3;         // return last node of condition block

        // prepare branch, fail and join blocks
        Block branch = new Block();
        Block fail = new Block();
        Block join = new Block( BlockKind.JOIN );
        if( curBlock.GetBranch() != null )
            join.SetBranch( curBlock.GetBranch() );
        curBlock.SetBranch( cond );
        cond.SetBranch( branch );
        cond.SetFail( fail );
        branch.SetBranch( join );
        fail.SetBranch( join );
        branch.SetLink( cond );
        fail.SetLink( cond );
        join.SetLink( cond );
        branch.SetDominator( cond );
        fail.SetDominator( cond );
        join.SetDominator( cond );

        // all blocks get dummy nodes as first nodes
        // this allows to link the jump nodes
        Node node4 = new Node( branch );
        node4.SetKind( NodeKind.NOP );
        node4.AddTail();
        node2.SetY( node4 );    // set jump target
        Node node5 = new Node( fail );
        node5.SetKind( NodeKind.NOP );
        node3.SetY( node5 );    // set jump target
        // Note: don't add fail node to node chain yet,
        // we first need to add branch nodes
        Node node6 = new Node( join );
        node6.SetKind( NodeKind.NOP );
        // Note: don't add join node to node chain yet,
        // we first need to add branch and fail nodes

        m_currentBlock = branch;
        return resultNode;
    }

    // jump from the end of the then (i.e. branch) block to the join block
    public Node Then( Node node )
    {
        Node resultNode = new Node( GetCurrentBlock() );
        resultNode.SetKind( NodeKind.INSTRUCTION );
        resultNode.SetY( GetBranchNode( resultNode ) );
        resultNode.SetOp( Operator.BRA );
		resultNode.AddTail();
        Node f = resultNode;
        if( resultNode.GetBlock().GetKind() == BlockKind.JOIN )
            f = GetBranchNode( f );
        f = GetLinkNode( f );
        m_currentBlock = f.GetBlock().GetFail();
        return resultNode;
    }

    public Node ElseInit( Node node )
    {
        Node resultNode = m_currentBlock.GetFirst();
        // add this node to the node list now that the then part is finished
        resultNode.AddTail();
        return resultNode;
    }

    // jump from the end of the else (ie. fail) block to the join block
    // also add first node of join block to nodelist since this is now
    // our current node
    public Node Else( Node node )
    {
        Node node1 = new Node( GetCurrentBlock() );
        node1.SetKind( NodeKind.INSTRUCTION );
        node1.SetY( GetBranchNode( node1 ) );
        node1.SetOp( Operator.BRA );
		node1.AddTail();
        Node resultNode = GetBranchNode( node1 );
        if( resultNode != null )
            resultNode.AddTail();
        m_currentBlock = resultNode.GetBlock();
        return resultNode;
    }

    public Node ElseExit( Node node )
    {
        Node resultNode = GetBranchNode( node );
        m_currentBlock = resultNode.GetBlock();
        if( resultNode != null )
            resultNode.AddTail();
        SavePhi( resultNode );
        return resultNode;
    }

    // comparison of two expressions, then prepare fail and join
    // blocks for while-do-end
    public Node While( Operator op, Node l, Node r )
    {
        Node resultNode = null;
        // create new condition block
        // Note: don't set branch of current block yet, we need it later
        Block cond = new Block( BlockKind.WHILE );
        Block curBlock = GetCurrentBlock();
        cond.SetLink( curBlock );
        cond.SetDominator( curBlock );

        Node node1 = new Node( cond );
        node1.SetKind( NodeKind.INSTRUCTION );
        node1.SetX( l );    // i < j etc., this is i
        node1.SetY( r );    // i < j etc., this is j
        node1.SetOp( Operator.CMP );
        node1.AddTail();
        Node node2 = new Node( cond );
        node2.SetKind( NodeKind.INSTRUCTION );
        node2.SetX( node1 );
        node2.SetOp( op );  // blt, etc to y, i.e. to "fail" block (while body)
        node2.AddTail();
        Node node3 = new Node( cond );
        node3.SetKind( NodeKind.INSTRUCTION );
        node3.SetOp( Operator.BRA );    // bra to y, i.e. to join block
        node3.AddTail();
        resultNode = node3;         // return last node of condition block

        // prepare fail and join blocks
        Block fail = new Block();
        Block join = new Block( BlockKind.JOIN );
        if( curBlock.GetBranch() != null )
            join.SetBranch( curBlock.GetBranch() );
        curBlock.SetBranch( cond );
        cond.SetBranch( join );
        cond.SetFail( fail );
        fail.SetBranch( cond );
        fail.SetLink( cond );
        join.SetLink( cond );
        join.SetDominator( cond );
        fail.SetDominator( cond );

        // all blocks get dummy nodes as first nodes
        // this allows to link the jump nodes
        Node node4 = new Node( fail );
        node4.SetKind( NodeKind.NOP );
        node4.AddTail();
        node2.SetY( node4 );    // set jump target
        Node node5 = new Node( join );
        node5.SetKind( NodeKind.NOP );
        node3.SetY( node5 );    // set jump target
        
        // Note: don't add join node to node chain yet,
        // we first need to add branch and fail nodes
        m_currentBlock = fail;
        return resultNode;
    }

    // jump from the end of the do (i.e. fail) block to the start block
    // also add first node of join block to nodelist since this is now
    // our current node
    public Node Do( Node node )
    {
        Node node1 = new Node( GetCurrentBlock() );
        node1.SetKind( NodeKind.INSTRUCTION );
        node1.SetY( GetBranchNode( node1 ) );
        node1.SetOp( Operator.BRA );
		node1.AddTail();
        Node resultNode = GetBranchNode( node1 );
        resultNode = GetBranchNode( resultNode );
        if( resultNode != null )
            resultNode.AddTail();
        m_currentBlock = resultNode.GetBlock();
        SavePhi( resultNode );
        node1.SetY( GetBranchNode( node1 ) );
        return resultNode;
    }

    // call procedure
    // only 3 procedures are supported: Read, Write, WriteLn
    public Node ProcedureCall( String id, Node p )
    {
        Node resultNode = new Node( GetCurrentBlock() );
        resultNode.SetKind( NodeKind.INSTRUCTION );
        if( id.equalsIgnoreCase( "Read" ) )
        {
            resultNode.SetOp( Operator.READ );
            resultNode.SetX( p );
            resultNode.SetValue( p.GetValue() );
            // for assignment 1 only
            if( m_Assignments == 1 )
            {
                resultNode.SetLastUse( (Node)m_symbolTable.get( resultNode.GetValue() ) );
                m_symbolTable.put( resultNode.GetValue(), resultNode );
            }
            else
            {
                // for assignment 2
                Node lastUse = (Node)m_symbolTable.get( resultNode.GetValue() );
                m_symbolTable.put( resultNode.GetValue(), resultNode );
                if( lastUse == null )
                    resultNode.SetSSANumber( 0 );
                else
                {
                    resultNode.SetLastUse( lastUse );
                    resultNode.SetSSANumber( lastUse.GetSSANumber() + 1 );
                }
                UpdatePhi( resultNode );
                // end change for assignment 2
            }
        }
        else if( id.equalsIgnoreCase( "Write" ) )
        {
            resultNode.SetOp( Operator.WRITE );
            resultNode.SetX( p );
            resultNode.SetValue( p.GetValue() );
            // for assignment 1 only
            if( m_Assignments == 1 )
            {
                resultNode.SetLastUse( (Node)m_symbolTable.get( resultNode.GetValue() ) );
                m_symbolTable.put( resultNode.GetValue(), resultNode );
            }
            else
            {
                // for assignment 2
                Node lastUse = (Node)m_symbolTable.get( resultNode.GetValue() );
                m_symbolTable.put( resultNode.GetValue(), resultNode );
                if( lastUse == null )
                    resultNode.SetSSANumber( 0 );
                else
                {
                    resultNode.SetLastUse( lastUse );
                    resultNode.SetSSANumber( lastUse.GetSSANumber() + 1 );
                }
                UpdatePhi( resultNode );
                // end change for assignment 2
            }
        }
        else if( id.equalsIgnoreCase( "WriteLn" ) )
            resultNode.SetOp( Operator.WLN );
        else 
        {
            // ignore all other procedure calls
            resultNode.SetKind( NodeKind.NOP );
        }
        resultNode.AddTail();
        return resultNode;
    }

    // the whole module is processed
    public Node Module( Node node )
    {
        Node resultNode = new Node( GetCurrentBlock() );
        resultNode.SetKind( NodeKind.INSTRUCTION );
        resultNode.SetOp( Operator.END );
        resultNode.AddTail();

        if( m_Assignments == 2 )
        {
            // for assignment 2
            if( m_ps != null )
            {
                m_ps.println();
                m_ps.println( "Common Subexpression Elimination" );
                m_ps.println( "================================" );
            }
            GlobalCSE( m_blockRoot, new Hashtable() );
            RebuildUseList();
            if( m_ps != null )
                m_ps.println();
            // end change for assignment 2
        }

        return resultNode;
    }


    // support functions

    // combine condition
    public Vector Condition( NodeKind kind, Node e1, Node e2 )
    {
        Vector cond = new Vector();
        if( kind == NodeKind.EQ )
            cond.addElement( Operator.BEQ );
        else if( kind == NodeKind.NE )
            cond.addElement( Operator.BNE );
        else if( kind == NodeKind.GT )
            cond.addElement( Operator.BGT );
        else if( kind == NodeKind.GE )
            cond.addElement( Operator.BGE );
        else if( kind == NodeKind.LT )
            cond.addElement( Operator.BLT );
        else if( kind == NodeKind.LE )
            cond.addElement( Operator.BLE );
        else
            cond.addElement( Operator.NOP );

        cond.addElement( e1 );
        cond.addElement( e2 );
        return cond;
    }

    protected void UpdatePhi( Node currentNode )
    {
        boolean updateX = false;
        Block curBlock = currentNode.GetBlock();
        Block link = curBlock.GetLink();
        if( link == null )
            return;
        while( link.GetKind() == BlockKind.WHILE && curBlock.GetKind() == BlockKind.JOIN )
        {
            curBlock = curBlock.GetBranch();
            link = curBlock.GetLink();
            if( link == null )
                return;
        }
        if( link.GetKind() == BlockKind.IF )
        {
            if( link.GetBranch() == curBlock )
                updateX = true;
            Block join = curBlock.GetBranch();
            if( join == null )
                return;
            Node node = join.GetFirst();
            FindPhi( node, currentNode, updateX );
        }
        else if( link.GetKind() == BlockKind.WHILE )
        {
            Node node = link.GetFirst();
            FindPhi( node, currentNode, false );
        }
    }

    protected void FindPhi( Node phiNode, Node currentNode, boolean updateX )
    {
        if( phiNode.GetKind() == NodeKind.NOP && phiNode.GetNext() != null )
            phiNode = phiNode.GetNext();
        Node node = phiNode;
        while( node != null && node.GetBlock() == phiNode.GetBlock() )
        {
            if( node.GetOp() == Operator.PHI )
            {
                if( node.GetValue().equals( currentNode.GetValue() ) )
                {
                    if( updateX )
                        node.SetX( currentNode );
                    else
                        node.SetY( currentNode );
                    return;
                }
                node = node.GetNext();
            }
            break;
        }

        Node phi = new Node();
        phi.SetBlock( phiNode.GetBlock() );
        phi.SetKind( NodeKind.INSTRUCTION );
        phi.SetOp( Operator.PHI );
        phi.SetValue( currentNode.GetValue() );
        if( updateX )
        {
            phi.SetX( currentNode );
            phi.SetY( currentNode.GetLastUse() );
        }
        else
        {
            phi.SetX( currentNode.GetLastUse() );
            phi.SetY( currentNode );
        }
        if( phiNode.GetKind() == NodeKind.NOP )
            phi.InsertAfter( phiNode, false );
        else
            phi.InsertBefore( phiNode, false );

        Node lastUse = (Node)m_symbolTable.get( phi.GetValue() );
        m_symbolTable.put( phi.GetValue(), phi );
        if( lastUse == null )
            phi.SetSSANumber( 0 );
        else
        {
            phi.SetLastUse( lastUse );
            phi.SetSSANumber( lastUse.GetSSANumber() + 1 );
        }

        if( m_firstPhi && m_ps != null )
        {
            m_ps.println( "Adding Phi instructions" );
            m_ps.println( "=======================" );
            m_firstPhi = false;
        }
        String outstr = "Add new Phi node: " + phi + ", X: " + phi.GetX() + ", Y: " + phi.GetY();
        if( m_ps != null )
            m_ps.println( outstr );
        else
            System.out.println( outstr );
    }

    protected void SavePhi( Node currentNode )
    {
        if( currentNode.GetKind() == NodeKind.NOP )
            currentNode = currentNode.GetNext();
        Node node = currentNode;
        while( node != null && node.GetBlock() == currentNode.GetBlock() )
        {
            if( node.GetOp() == Operator.PHI )
            {
                Node lastUse = (Node)m_symbolTable.get( node.GetValue() );
                if( lastUse == null )
                    node.SetSSANumber( 0 );
                else
                {
                    node.SetLastUse( lastUse );
                    node.SetSSANumber( lastUse.GetSSANumber() + 1 );
                }
                node = node.GetNext();
            }
            break;
        }
    }

	protected void GlobalCSE( Block block, Hashtable anchor )
	{
        String outstr = "Global CSE, current block: " + block;
        if( m_ps != null )
            m_ps.println( outstr );
        else
            System.out.println( outstr );
		LocalCSE( block, anchor );
		Hashtable anchorClone = (Hashtable)anchor.clone();
		Block child = block.GetChildDominator();
		if( child != null )
			GlobalCSE( child, anchorClone );
//		else
//		{
			Block sibling = block.GetSiblingDominator();
			if( sibling != null )
				GlobalCSE( sibling, anchorClone );
//		}
	}

	protected void LocalCSE( Block block, Hashtable anchor )
	{
        String outstr = "Local CSE, current block: " + block;
        if( m_ps != null )
            m_ps.println( outstr );
        else
            System.out.println( outstr );
		Node node = block.GetFirst();
		anchor.remove( Operator.PHI.toString() );
		while( node != null && node.GetBlock() == block )
		{
			if( node.GetOp() == Operator.PHI )
			{
                // Note: if we ever have more than 2 parameters
                // for phi then we need to keep them in a linked
                // list under X or Y
                // We then would need to walk the list here
                node.SetX( EquivalentOperand( node.GetX() ) );
                node.SetY( EquivalentOperand( node.GetY() ) );

                Node lastAnchor = (Node)anchor.get( node.GetOp().toString() );
				Node equiv = EquivalentInstruction( node.GetX(), node.GetY(), lastAnchor );
				if( equiv != null )
				{
                    outstr = "Erasing node: " + node;
                    if( m_ps != null )
                        m_ps.println( outstr );
                    else
                        System.out.println( outstr );
					node.SetX( equiv );
					node.SetErased();
				}
				else
				{
					node.SetOpLink( lastAnchor );
					anchor.put( node.GetOp().toString(), node );
				}
			}
			else if( node.GetOp() == Operator.MOVE )
			{
                outstr = "Erasing node: " + node;
                if( m_ps != null )
                    m_ps.println( outstr );
                else
                    System.out.println( outstr );
				node.SetX( EquivalentOperand( node.GetX() ) );
				node.SetErased();
			}
			else if( node.GetOp() != null )
			{
				node.SetX( EquivalentOperand( node.GetX() ) );
				node.SetY( EquivalentOperand( node.GetY() ) );
				Node lastAnchor = (Node)anchor.get( node.GetOp().toString() );
				Node equiv = EquivalentInstruction( node.GetX(), node.GetY(), lastAnchor );
				if( equiv != null )
				{
                    outstr = "Erasing node: " + node;
                    if( m_ps != null )
                        m_ps.println( outstr );
                    else
                        System.out.println( outstr );
					node.SetX( equiv );
					node.SetErased();
				}
				else
				{
                    outstr = "Linking anchor: " + node + ", last anchor: " + lastAnchor;
                    if( m_ps != null )
                        m_ps.println( outstr );
                    else
                        System.out.println( outstr );
					node.SetOpLink( lastAnchor );
					anchor.put( node.GetOp().toString(), node );
				}
			}
            node = node.GetNext();
		}
	}

	protected Node EquivalentOperand( Node node )
	{
		if( node != null && node.IsErased() )
			return node.GetX();
		else
			return node;
	}

	protected Node EquivalentInstruction( Node x, Node y, Node last )
	{
		while( last != null )
		{
			if( last.GetOp() == Operator.PHI && !last.IsErased() )
			{
                while( last != null )
                {
                    // Note: if we ever have more than 2 parameters
                    // for phi then we need to keep them in a linked
                    // list under X or Y
                    // We then would need to walk the list here
                    Node p = x;
                    Node q = last.GetX();
                    if( p != null && q != null && p.GetValue().equals( q.GetValue() ) )
                    {
                        p = y;
                        q = last.GetY();
                        if( p != null && q != null && p.GetValue().equals( q.GetValue() ) )
                            p = q = null;
                    }
                    if( p == null && q == null )
                        return last;
                    else
        				last = last.GetOpLink();
                }
			}
			else if( last.GetX() == x && last.GetY() == y )
				return last;
			else
				last = last.GetOpLink();
		}

		return null;
	}

    public void RebuildUseList()
    {
        m_symbolTable.clear();

        Node node = m_nodeRoot;
        while( node != null )
        {
            if( !node.IsErased() )
            {
                if( node.GetKind() == NodeKind.ID )
                {
                    if( node.GetlValue() )
                    {
		                node.SetLastUse( (Node)m_symbolTable.get( node.GetValue() ) );
    		            m_symbolTable.put( node.GetValue(), node );
                    }
                    else
                    {
                        Node lastNode = (Node)m_symbolTable.get( node.GetValue() );
		                if( lastNode == null )
        		            m_symbolTable.put( node.GetValue(), node );
                        else
                            node.SetLastUse( lastNode );
                    }
                }
                else if( node.GetKind() == NodeKind.INTEGER )
                {
                    Node lastNode = (Node)m_symbolTable.get( node.GetValue() );
		            if( lastNode == null )
			            m_symbolTable.put( node.toString(), node );
                    else
                        node.SetLastUse( lastNode );
                }
                else if( node.IsValueSet() )
                {
		            node.SetLastUse( (Node)m_symbolTable.get( node.GetValue() ) );
                    m_symbolTable.put( node.GetValue(), node );
                }

            }
            node = node.GetNext();
        }
    }

    // return the first node of the branch block of a node if there is one
    public Node GetBranchNode( Node node )
    {
        Block b = node.GetBlock();
        if( b != null )
            b = b.GetBranch();
        if( b != null )
            return b.GetFirst();
        else
            return null;
    }

    // return the first node of the fail block of a node if there is one
    public Node GetFailNode( Node node )
    {
        Block b = node.GetBlock();
        if( b != null )
            b = b.GetFail();
        if( b != null )
            return b.GetFirst();
        else
            return null;
    }

    // return the first node of the link block of a node if there is one
    public Node GetLinkNode( Node node )
    {
        Block b = node.GetBlock();
        if( b != null )
            b = b.GetLink();
        if( b != null )
            return b.GetFirst();
        else
            return null;
    }

    public String TraverseNodes()
    {
        if( m_nodeRoot != null )
        {
            Node node = m_nodeRoot;
            if( node != null )
                return node.Traverse();
        }
        return "No Nodes\n";
    }

    public String TraverseBlocks()
    {
        if( m_blockRoot != null )
            return m_blockRoot.Traverse();
        return "No Parse Tree\n";
    }

    public String TraverseDominatorTree()
    {
        if( m_blockRoot != null )
            return m_blockRoot.TraverseDominatorTree();
        return "No Dominator Tree\n";
    }

    public String TraverseUselist()
    {
        String str = "Uselist\n";
        str +=       "=======\n";
        Enumeration e = m_symbolTable.keys();
        while( e.hasMoreElements() )
        {
            String key = (String)e.nextElement();
            Node node = (Node)m_symbolTable.get( key );
            int ind = 0;
            while( node != null && !node.IsVisited() )
            {
                node.Visited();
                str += indent( ind ) + node + " Value: " + node.GetValue() + "\n";
                ind += 2;
                node = node.GetLastUse();
            }
        }
        return str;
    }

    // add some predefined nodes
    public void InitNodes()
    {
        Node node = Ident( "FP", false );  // base for array calculations
        if( m_nodeRoot == null )
            m_nodeRoot = node;
        node = Ident( "INTEGER", false );
    }

    public Block StartBlock()
    {
        Block newBlock = new Block();
        if( m_blockRoot == null )
        {
            m_blockRoot = newBlock;
            newBlock.SetFirst( m_nodeRoot );
            Node node = m_nodeRoot;
            while( node != null )
            {
                node.SetBlock( newBlock );
                node = node.GetNext();
            }
        }
        m_currentBlock = newBlock;
        newBlock.AddTail();
        return newBlock;
    }

    public Block GetCurrentBlock()
    {
        return m_currentBlock;
    }

    public Node GetCurrentNode()
    {
        return Node.endNode;
    }

    private String indent( int ind )
	{
		String str = "";
		for( int i=0; i<ind; i++ )
			str += " ";
		return str;
	}
}

